/*
 *	Added functionality to modify espeak parameters and espeak_VOICE Parameters through D-Bus methods
 *	Added feature to write out a configuration file, and automatically load it on service startup
 */
/*
 *  Modified the Load cofiguration function to load default or user configuration file
 */
/* Function accessible VIA python interface
 * 
 * 
 * import dbus
 * bus = dbus.SessionBus()
 * espeak_object = bus.get_object('org.laptop.Speech','/org/laptop/Speech')
 * Above three lines initialize the dbus service for espeak and create an espeak object for it.
 * 
 * 
 * espeak_object.SetVoice(string language,int gender,int age)
 * 
 * espeak_object.SayText(string text)
 * 
 * espeak_object.SetParameter(int PARAMETER_NAME, int PARAMETER_VALUE)
 * 
 * espeak_object.ShowConfiguration()
 * 
 * espeak_object.LoadConfiguration(int default_flag) 
 * 
 * espeak_object.SaveConfiguration()
 */


//TODO: Write finalizer method to deallocate the the memory allocated to get_speech_parameters.
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <glib.h>
#include <glib-object.h>
#include <gmodule.h>

#include <dbus/dbus-protocol.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include <espeak/speak_lib.h>

#include "speech-server.h"

#define MAX_STR_SIZE			2000

#define	ERR_INVALID_STR			0
#define ERR_SPEECH_FAIL			1
#define ERR_PARAMETER_SET_FAIL	2
#define ERR_VOICE_SET_FAIL		3
#define ERR_PARAMETER_SAVE		4
#define MAX_STR_SIZE			2000

#define ERR_QUARK()		_speech_server_err_quark	

static GQuark _speech_server_err_quark = 0;



static GEnumValue _speech_server_errors[] =
{
	{ ERR_INVALID_STR,		"Invalid string",		"InvalidString" },
	{ ERR_SPEECH_FAIL,		"Synthesis failure",		"SynthesisFailure" },
	{ ERR_PARAMETER_SET_FAIL,	"Speech Parameter Not Set",	"ParameterSetFailure" },
	{ ERR_VOICE_SET_FAIL,		"Voice Parameters Not Set",	"VoiceSetFailure" },
	{ ERR_PARAMETER_SAVE,		"Espeak Parameters Not Saved"	"ParamaterSaveFailure"},
	{ 0, NULL, NULL }
};

static char * _aplay_cmdline = NULL;
static char * const configuration_file = "data/speech_server_configuration.out";
static char * const default_configuration_file = "data/speech_server_default_configuration.out";


static int _synth_func(short * data, int len, espeak_EVENT * evt)
{
	fwrite((void *)data, sizeof(short), len, (FILE *)evt->user_data);
	return EE_OK;
}

static gboolean speech_server_say_text(SpeechServer * speech_server,
									   const char * text,
									   GError **error);
/* 
 * espeak_object.SayText(string text)//As called from python interface
 * Speakes out the text pointed by string in the voice set by SetVoice() of in default language.
 */

static gboolean speech_server_set_parameter(SpeechServer * speech_server,
											gint parameter_name,
											gint parameter_value,
											GError **error);
/* 
 * espeak_object.SetParameter(int PARAMETER_NAME, int PARAMETER_VALUE)//As called from python intertface
 * This sets the value of PARAMETER_NAME to PARAMETER_VALUE
 * espeak_PARAMETERS and thier equivalent integer value->
 *	 espeakRATE = 1
 *   espeakVOLUME = 2
 *   espeakPITCH = 3
 *   espeakRANGE = 4
 *   espeakPUNCTUATION = 5
 *   espeakCAPITALS = 6
 */


static gboolean speech_server_set_voice(SpeechServer * speech_server,
										gchar *name,
										gchar *languages,
										gchar *identifier,
										gint gender,
										gint age,
										gint variant,
										GError **error);
/* 
 * espeak_object.SetVoice(string language, gint gender,gint age)//As called from python interface
 * This functin sets the voice for language as language(In standard notation)
 * 								gender as gender( 1 = male, 2 = female, 0 = unknown)
 * 								age as age
 * 
 */


static gboolean speech_server_save_configuration(SpeechServer * speech_server,
												 GError **error);
/* 
 * espeak_object.SaveConfiguration()//As called from python interface
 * This saves the current configuration of espeak to a file which can be later load using LoadConfiguration(0)
 */

static gboolean speech_server_get_configuration(SpeechServer * speech_server,
												GValueArray **speech_parameters_array,
												GError **error);
/* 
 * espeak_object.ShowConfiguration()//As called from python interface
 * This currently displays the value of all the parameters on server side
 */


static void getParameters(SpeechServerParameters *);

static gboolean speech_server_load_configuration(const char *conf_path);
static gboolean speech_server_load_default_configuration(SpeechServer * speech_server,
														 GError **error);
static gboolean speech_server_load_user_configuration(SpeechServer * speech_server,
												 GError **error);

static void speech_server_print_configuration(SpeechServerParameters *speech_parameters);
/* 
 * espeak_object.LoadConfiguration(int default_flag) //As called from python interface
 * This load the espeak configuration from a file.
 * 	if default_flag = 0 -> loads the previously saved configuration (saved using SaveConfiguration)
 * 	if default_flag = 1 -> loads the default saved configuration
 */

//dbus-binding-tool --mode=glib-server --prefix=speech_server speech-server.xml > speech-server-glue.h
#include "speech-server-glue.h"

static void speech_server_class_init (SpeechServerClass * klass)
{
	dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (klass),
									 &dbus_glib_speech_server_object_info);
	
	_speech_server_err_quark = g_quark_from_static_string("speech_server_err_quark");
	
	dbus_g_error_domain_register(_speech_server_err_quark,
								 "org.laptop.Speech",
								 g_enum_register_static("SpeechError",
														_speech_server_errors));
}

static void speech_server_init (SpeechServer * speech_server)
{
	GError **error;

	int ret = espeak_Initialize(AUDIO_OUTPUT_SYNCHRONOUS, 0, NULL, 0);
	
	
	espeak_SetSynthCallback(_synth_func);
	if(ret != EE_INTERNAL_ERROR)
		_aplay_cmdline = g_strdup_printf("aplay -q -f S16_LE -r %d -", ret);
    
	//Load the personalized configuration for the user 
	speech_server_load_user_configuration(speech_server, error);	
}

//G_DEFINE_TYPE(SpeechServer, speech_server, G_TYPE_OBJECT);

static gpointer speech_server_parent_class = NULL;

static void speech_server_class_intern_init (gpointer klass) 
{
	speech_server_parent_class = g_type_class_peek_parent (klass); 
	speech_server_class_init ((SpeechServerClass*) klass); 
}

GType speech_server_get_type()
{
	static GType _type = 0;
	if(_type == 0)
		_type = g_type_register_static_simple(G_TYPE_OBJECT, "SpeechServer", 
			sizeof(SpeechServerClass),  (GClassInitFunc)speech_server_class_intern_init,
			sizeof(SpeechServer),  (GInstanceInitFunc)speech_server_init, 0);
	return _type;
}

static gboolean speech_server_say_text(SpeechServer * speech_server,
				       const char * text,
				       GError ** error)
{
	int str_len;
	FILE * f;

	if(_aplay_cmdline == NULL)
	{
		* error = g_error_new(ERR_QUARK(),ERR_SPEECH_FAIL,
				      "ESpeak failed to initialize the synthesis engine.");
		return FALSE;	
	}

	f = popen(_aplay_cmdline, "w");

	if(! f)
	{
		* error = g_error_new(ERR_QUARK(),ERR_SPEECH_FAIL,
				      "ESpeak failed to exec 'aplay'. [%s]", strerror(errno));
		return FALSE;	
	}

	str_len = strlen(text);

	if(str_len > MAX_STR_SIZE)
	{
		* error = g_error_new(ERR_QUARK(),ERR_INVALID_STR,
				      "ESpeak received a string larger than %d bytes.",
				      MAX_STR_SIZE);		
		return FALSE;
	}

	if(! g_utf8_validate(text, str_len, NULL))
	{
		* error = g_error_new(ERR_QUARK(),ERR_INVALID_STR,
				      "ESpeak received an invalid UTF8 string.");		
		return FALSE;
	}
	
	if(espeak_Synth(text, str_len, 0, POS_CHARACTER, 0, espeakCHARS_UTF8, 
		NULL, (void *)f) != EE_OK)
	{
		* error = g_error_new(ERR_QUARK(),ERR_SPEECH_FAIL,
				      "ESpeak failed to synthesize the string \"%s\".",
				      text);
		return FALSE;
	}
	
	fclose(f);
	return TRUE;
}

static gboolean speech_server_set_parameter(SpeechServer * speech_server,
					    					gint  parameter_name,
					    					gint parameter_value,
										    GError **error)
{
	if(espeak_SetParameter((espeak_PARAMETER) parameter_name, parameter_value, 0) != EE_OK)
	{
		*error = g_error_new(ERR_QUARK(),ERR_PARAMETER_SET_FAIL,
				     "Espeak failed to set the parameter value");
		return FALSE;
	}	
	
	return TRUE;
}

static gboolean speech_server_set_voice(SpeechServer * speech_server,
										gchar *name,
										gchar *languages,
										gchar *identifier,
										gint gender,	
										gint age,
										gint variant,
										GError **error)
{
	
	espeak_VOICE voice_parameters;
		
	voice_parameters.name  = name;
	voice_parameters.languages = languages;
	voice_parameters.identifier = identifier;
	voice_parameters.gender = (guchar) gender;
	voice_parameters.age = (guchar) age;
	voice_parameters.variant = (guchar) variant;
	
	//somehow getparameters() is not returning the correct values for these
	//parameters! So we'll maintain them ourselves in the global SpeechServerParameters
	//this is a crazy idea, and we'll have to fix it soon.
	//strcpy(speech_parameters.languages, languages);
	//speech_parameters.age = (guchar) age;
	//speech_parameters.gender = (guchar) gender;
	
	if(espeak_SetVoiceByProperties(&voice_parameters) != EE_OK)
	{
		*error = g_error_new(ERR_QUARK(),ERR_VOICE_SET_FAIL,
				     "Espeak failed to set the voice parameters");
		return FALSE;
	}	
	
	return TRUE;
}


static void getParameters(SpeechServerParameters *speech_parameter)
{
	
	espeak_VOICE *voice_parameters;

		
	speech_parameter->rate			=	espeak_GetParameter(espeakRATE, 1);
	speech_parameter->volume		=	espeak_GetParameter(espeakVOLUME, 1);
	speech_parameter->pitch			=	espeak_GetParameter(espeakPITCH, 1);
	speech_parameter->range			=	espeak_GetParameter(espeakRANGE, 1);
	speech_parameter->punctuation	=	espeak_GetParameter(espeakPUNCTUATION, 1);
	speech_parameter->capitals		=	espeak_GetParameter(espeakCAPITALS, 1);
	
	voice_parameters = espeak_GetCurrentVoice();
	
	strcpy(speech_parameter->languages,voice_parameters->languages);
    strcpy(speech_parameter->name,voice_parameters->name);
	strcpy(speech_parameter->identifier,voice_parameters->identifier);
	
	speech_parameter->gender 		=	(guchar) 	voice_parameters->gender;
	speech_parameter->age 			=	(guchar) 	voice_parameters->age;	
	speech_parameter->variant 		=	(guchar) 	voice_parameters->variant;
	
	if(debug)
		speech_server_print_configuration(speech_parameter);
}


static gboolean speech_server_save_configuration(SpeechServer * speech_server, GError **error)
{
	//Get the structure with all the parameter properties
	SpeechServerParameters speech_parameters;
	
	getParameters(&speech_parameters);
	  
	FILE *out_file;
	int write_size;	
	
	out_file = fopen(configuration_file,"wb");
	
	if( out_file == NULL )
	{
	    *error = g_error_new(ERR_QUARK(),ERR_PARAMETER_SAVE,
				 "Espeak failed to save the present settings!");
	    return FALSE;
	}
	
	write_size = fwrite((char*)(&speech_parameters), 1,
						sizeof(SpeechServerParameters), out_file);
	if(write_size < 0)
	{
	    *error = g_error_new(ERR_QUARK(),ERR_PARAMETER_SAVE,
							 "Espeak failed to save the present settings!");
	    return FALSE;
	}

	fclose(out_file);   
    
    	
	return TRUE;
}

static gboolean speech_server_load_configuration(const char * conf_path)
{
   	FILE* in_file;
  	int read_size;
	espeak_VOICE voice_parameters;
	SpeechServerParameters speech_saved_parameters;
	
	in_file = fopen(conf_path,"rb");    	
	 	
    
    if( in_file == NULL ){
		perror("speech_server::speech_server_load_configuration::");
		return FALSE;	    
	}
	    
	read_size = fread((char*)(&speech_saved_parameters), 1,
					  sizeof(SpeechServerParameters), in_file);
    if(read_size < 0){
		perror("speech_server::speech_server_load_configuration::");
		return FALSE;
	}
	
    fclose(in_file);   	
	
		  	
	voice_parameters.name 		=	(char *)		speech_saved_parameters.name;
	voice_parameters.languages 	= 	(char *) 		speech_saved_parameters.languages;
	voice_parameters.identifier = 	(char *) 		speech_saved_parameters.identifier;
	voice_parameters.gender 	= 	(unsigned char)	speech_saved_parameters.gender;
	voice_parameters.age 		= 	(unsigned char)	speech_saved_parameters.age;
	voice_parameters.variant 	= 	(unsigned char)	speech_saved_parameters.variant;
	
	if(debug){
		printf("\nspeech_server::load_configuration::%s, %s, %s\n",
			   voice_parameters.name,
			   voice_parameters.languages,
			   voice_parameters.identifier);
		printf("\nspeech_server::load_configuration::Gender::%u\n",
			   voice_parameters.gender);
		printf("\nspeech_server::load_configuration::Age::%u\n",
			   voice_parameters.age);
		printf("\nspeech_server::load_configuration::Variant::%u\n",
			   voice_parameters.variant);
	}
    
	//Start setting the parameters now!
	espeak_SetParameter(espeakRATE, 		(int) speech_saved_parameters.rate, 0);
	espeak_SetParameter(espeakVOLUME, 		(int) speech_saved_parameters.volume, 0);
	espeak_SetParameter(espeakPITCH, 		(int) speech_saved_parameters.pitch, 0);
	espeak_SetParameter(espeakRANGE, 		(int) speech_saved_parameters.range, 0);
	espeak_SetParameter(espeakPUNCTUATION, 	(int) speech_saved_parameters.punctuation, 0);
	espeak_SetParameter(espeakCAPITALS, 	(int) speech_saved_parameters.capitals, 0);
	
    if(espeak_SetVoiceByProperties(&voice_parameters) != EE_OK)
		return FALSE;
		
	if(debug)
		speech_server_print_configuration(&speech_saved_parameters);
	
	return TRUE;
}

static gboolean speech_server_get_configuration(SpeechServer * speech_server,
												GValueArray **speech_parameters_array,
												GError **error)
{	
	
	
	SpeechServerParameters *get_speech_parameters;	
	get_speech_parameters = (SpeechServerParameters *) g_try_malloc(sizeof(SpeechServerParameters));
	if(get_speech_parameters == NULL){
		return FALSE;
		//TODO:Set Gerror Quark
	}
	
	getParameters(get_speech_parameters);
	
	//TODO::Analyze how speech_parameters_array grows in its lifetime, and wheter it hogs memory
	//TODO: Fix the hardcoded number 12
	*speech_parameters_array = g_value_array_new (12);
		
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->rate);
		
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->volume);
	
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->pitch);
	
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->range);
	
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->punctuation);
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->capitals);
				
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_STRING);
	g_value_set_string (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->name);
	
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_STRING);
	g_value_set_string (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->identifier);
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_STRING);
	g_value_set_string (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->languages);
	
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->gender);
	
			
	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->age);

	g_value_array_prepend (*speech_parameters_array, NULL);		
	g_value_init (g_value_array_get_nth (*speech_parameters_array, 0), G_TYPE_INT);
	g_value_set_int (g_value_array_get_nth (*speech_parameters_array, 0), get_speech_parameters->variant);

	g_free(get_speech_parameters);
	return TRUE;

}

static void speech_server_print_configuration(SpeechServerParameters *speech_parameters){
	
	printf("\nspeech_server::print_configuration::start::SPEECH SERVER PARAMETERS::\n");
	printf("Rate::%d\n", speech_parameters->rate);
	printf("Volume::%d\n", speech_parameters->volume);
	printf("Pitch::%d\n", speech_parameters->pitch);
	printf("Range::%d\n", speech_parameters->range);
	printf("Punctuation::%d\n", speech_parameters->punctuation);
	printf("Capitals::%d\n", speech_parameters->capitals);
	printf("VOICE PARAMETERS::\n");
	printf("Name::%s\n", speech_parameters->name);
	printf("Identifier::%s\n", speech_parameters->identifier);
	printf("Language::%s\n", speech_parameters->languages);
	printf("Gender::%u\n", speech_parameters->gender);
	printf("Age::%u\n", speech_parameters->age);
	printf("Variant::%u\n", speech_parameters->variant);
	printf("speech_server::print_configuration::end\n");
	return;	
}

static gboolean speech_server_load_user_configuration(SpeechServer * speech_server,
													  GError **error)
{
	if(speech_server_load_configuration(configuration_file) == TRUE)
		return TRUE;
	else{
		*error = g_error_new(ERR_QUARK(),
							 ERR_VOICE_SET_FAIL,
							 "Espeak failed to set the voice parameters");
		return FALSE;
	}
	
}

static gboolean speech_server_load_default_configuration(SpeechServer * speech_server,
												 GError **error)
{
	if(speech_server_load_configuration(default_configuration_file) == TRUE)
		return TRUE;
	else{
		*error = g_error_new(ERR_QUARK(),
							 ERR_VOICE_SET_FAIL,
							 "Espeak failed to set the voice parameters");
		return FALSE;
	}
	
}
